/* FILE: mesh.h                 -*-Mode: c++-*-
 *
 * Abstract mesh class and children
 *
 */

#ifndef MESH_H
#define MESH_H

#include "misc.h"
#include "ptsearch.h"
#include "template.h"
#include "vector.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// Mesh (abstract) class
class Mesh
{
private:
  const char * const MeshTypename;
  char *Filename;
  char *Title;
  char *Description;
protected:
  void SetFilename(const char* _Filename);
  void SetTitle(const char* _Title);
  void SetDescription(const char* _Description);
public:
  const char* GetName() const;
  const char* GetTitle() const;
  const char* GetDescription() const;
  // NOTE: "Filename", "Title" & "Description" are owned and managed
  // by class Mesh, and the returned address will become invalid if
  // SetFilename/Title/Descripion is called or if the mesh is
  // destroyed.  Calling routines will want to copy the returned
  // string into local memory if it is needed for an extended time.

  const char* GetMeshType() const { return MeshTypename; }

  virtual INT4m GetRange(BoundingBox &range) const =0;
  /// Data range.  Tightness not required, but should
  /// be close.

  virtual INT4m GetBoundaryList(List<Vec3f> &boundary_list) const =0;
  /// Used to draw outline polygon in display.  Should probably enclose
  /// all data points.

  virtual REAL4m GetApproximateCellWidth(void) const =0;
  virtual Vec3f  GetApproximateCellDimensions(void) const =0;
  /// This should be an "average" distance between nearest
  /// neighbor points.  Used to size arrows in display.

  virtual REAL4m GetPreferredStepSize(REAL4m step) const =0;
  /// Import "step" and return value is in mesh units.  For regular grids
  /// the return value should probably be an integral multiple of
  /// "GetApproximateCellWidth()".

  virtual REAL4m GetSubsampleGrit(void) const =0;
  /// Used to set increment on subsample scale widget in display control
  /// window.  Units are RELATIVE to "GetApproximateCellWidth()".  For
  /// regular grids it doesn't make sense to sample at a rate smaller
  /// than the grid lattice spacing, so this value should be 1 (provided
  /// GetApproximateCellWidth() returns the lattice spacing constant),
  /// but for irregular meshes, a smaller value may make perfect sense.
  ///  GetSubsampleGrit() should return a value >0 and <=1.

  virtual INT4m GetDisplayList(REAL4m &xstep_request,
			       REAL4m &ystep_request,
			       REAL4m &zstep_request,
			       const BoundingBox &range,
			       List<LocatedVector> &display_list) =0;
  ///   Use ?step_request=0 to remove constraint on that axis, e.g., if
  /// xstep_request=ystep_request=zstep_request=0, then this routine places
  /// all points inside "range" into display_list.
  ///   Return value is number of points in display_list.  This is conceptually
  /// a const function, but isn't if the underlying representation stores the
  /// mesh values internally in a List<> structure, and uses the List<>
  /// iterator to access the elements (the iterator has an internal pointer
  /// that gets incremented with each access).  One might want to add a List<>
  /// "header" copy routine to the List<> template definition, to allow for
  /// efficient local iterators.  Then this function could be made const.

  Mesh(const char *_MeshTypename,const char *_Filename,
       const char *_Title=NULL,const char* _Description=NULL):
    MeshTypename(_MeshTypename),Filename(NULL),
    Title(NULL),Description(NULL) {
      SetFilename(_Filename);
      SetTitle(_Title);
      SetDescription(_Description);
  }
  virtual INT4m GetSize() const =0;

  virtual ~Mesh() {
    if(Filename!=NULL)    delete[] Filename;
    if(Title!=NULL)       delete[] Title;
    if(Description!=NULL) delete[] Description;
  }
};

//////////////////////////////////////////////////////////////////////////
// Empty mesh
class EmptyMesh:public Mesh
{
private:
  BoundingBox empty_range;
  List<Vec3f> empty_boundary_list;
public:
  EmptyMesh():Mesh("EmptyMesh","<empty>") {}
  INT4m GetRange(BoundingBox &range) const { range=empty_range; return 0; }
  INT4m GetBoundaryList(List<Vec3f> &boundary_list) const {
    boundary_list=empty_boundary_list;
    return 0;
  }
  REAL4m GetApproximateCellWidth() const { return 1.; }
  Vec3f  GetApproximateCellDimensions() const { return Vec3f(1.,1.,1.); }
  REAL4m GetPreferredStepSize(REAL4m step) const { return rint(step); }
  REAL4m GetSubsampleGrit(void) const { return 1.; }
  INT4m GetDisplayList(REAL4m &,REAL4m &,REAL4m &,
		       const BoundingBox&,
		       List<LocatedVector> &display_list) {
    display_list.Clear();
    return 0;
  }
  INT4m GetSize() const { return 0; }
};

//////////////////////////////////////////////////////////////////////////
// Regular 3D grid of Vec3f's
class GridVec3f:public Mesh
{ // NOTE: dim1,dim2,dim3 of grid should be thought of in the order z,y,x
  // (or k,j,i), because in applications typically z-dim (k) will be smallest,
  // x-dim (i) will be largest, so the data is stored arr[k][j][i].  This is
  // why we overload the () operator to map (i,j,k) -> [k][j][i].
  // IN PARTICULAR, efficient loops will increment the innermost loop on i.
  //   SO, KEEP THIS STRAIGHT: (i,j,k) coming in to member functions
  // should access grid[k][j][i], or (better) use GridVec(i,j,k).
private:
  static const ClassDoc class_doc;
  Array3D<Vec3f> grid;     // Magnetization vector array
  inline INT4m SetSize(INT4m newi,INT4m newj,INT4m newk) {
    return grid.Allocate(newk,newj,newi);
  }
  BoundingBox range;
  List<Vec3f> boundary;
  Vec3f coords_base,coords_step;  
  Vec3f NodalToProblemCoords(Vec3f const &w_nodal) const;
  Vec3f ProblemToNodalCoords(Vec3f const &v_problem) const;
  /// The conversion from nodes to problem coordinates is
  ///    (i,j,k) -> base+(i*step.x,j*step.y,k*step.z).
  /// NOTE 1: The grid vector access functions, e.g., operator(), access
  ///   via the integral "nodal" coordinates, *not* in terms of the
  ///   (real valued) problem coordinates.
  /// NOTE 2: The values in coords_step are guaranteed to be !=0,
  ///   but may be negative!  However, GetStdStep() (below), returns
  ///   a value >0.
  /// NOTE 3: The coords conversion routines don't preserve order
  ///   if step.? is negative, e.g., say (min.x,min.y) < (max.x,max.y)
  ///   going in, but step.x is negative.  Then after conversion,
  ///   min.y~ will still be < max.y~, but now min.x~>max.x~.

  REAL4m GetStdStep() const {
    REAL4m basestep=MIN(fabs(coords_step.x),fabs(coords_step.y));
    return basestep;
  }

  // The following function is left undefined on purpose
  GridVec3f& operator=(GridVec3f& rhs); // Assignment op

public:
  inline INT4m GetSize(INT4m &isize,INT4m &jsize,INT4m &ksize) const {
    return grid.GetSize(ksize,jsize,isize);
  }
  INT4m GetSize() const { return grid.GetSize(); }
  GridVec3f():Mesh("GridVec3f",NULL,NULL) {}

  // The following constructor should be called from file input functions.
  GridVec3f(const char *filename,const char *title,const char *desc,
	    INT4m newi,INT4m newj,INT4m newk,
	    const Vec3f &basept,const Vec3f &gridstep,
	    const BoundingBox &newrange,const List<Vec3f> &newboundary);

  Vec3f GetBasePoint() const { return coords_base; }
  Vec3f GetGridStep() const { return coords_step; }

  GridVec3f(GridVec3f &other):Mesh(NULL,"GridVec3f") { // Copy constructor
    SetFilename(other.GetName());
    SetTitle(other.GetTitle());
    SetDescription(other.GetDescription());
    grid.Copy(other.grid);
    range=other.range;
    boundary=other.boundary;
    coords_base=other.coords_base;
    coords_step=other.coords_step;
  }

  inline Vec3f& GridVec(INT4m i,INT4m j,INT4m k) { return grid(k,j,i); }
  inline Vec3f& operator()(INT4m i,INT4m j,INT4m k) { return GridVec(i,j,k); }
  inline const Vec3f& GridVec(INT4m i,INT4m j,INT4m k) const {
    return grid(k,j,i);
  }
  inline const Vec3f& operator()(INT4m i,INT4m j,INT4m k) const {
    return GridVec(i,j,k);
  }
  virtual INT4m GetRange(BoundingBox &myrange) const {
    myrange=range; return 0;
  }
  virtual INT4m GetBoundaryList(List<Vec3f> &boundary_list) const {
    boundary_list=boundary;
    return 0;
  }
  virtual REAL4m GetApproximateCellWidth() const { return GetStdStep(); }
  Vec3f  GetApproximateCellDimensions() const {
    return Vec3f(fabs(coords_step.x),fabs(coords_step.y),
		 fabs(coords_step.z));
  }
  REAL4m GetPreferredStepSize(REAL4m step) const {
    REAL4m basestep=GetStdStep();
    return basestep*rint(step/basestep);
  }
  REAL4m GetSubsampleGrit(void) const { return 1.; }
  virtual INT4m GetDisplayList(REAL4m &xstep_request,
			       REAL4m &ystep_request,
			       REAL4m &zstep_request,
			       const BoundingBox &dsprange,
			       List<LocatedVector> &display_list);
  virtual ~GridVec3f(void) {}
};

//////////////////////////////////////////////////////////////////////////
// General (non-regular) 3D mesh of Vec3f's
class GeneralMesh3f:public Mesh
{ 
private:
  static const ClassDoc class_doc;
  BoxList mesh;
  BoundingBox data_range;  // Minimal box containing data points
  BoundingBox range;       // Includes data_range+boundary
  List<Vec3f> boundary;    // 2D boundary vertices
  Vec3f step_hints;        // Hints on grid structure

  // The following function is left undefined on purpose
  GeneralMesh3f& operator=(GeneralMesh3f& rhs); // Assignment op

  // Provided for convience for mesh arrays.
  GeneralMesh3f(void):Mesh("GeneralMesh3f",NULL) {}
public:
  ////////////// Generic Mesh functions ///////////////////
  GeneralMesh3f(const char* filename,const char* title=NULL,
		const char* desc=NULL)
    :Mesh("GeneralMesh3f",filename,title,desc) {}
  /// The preceding constructor should be called from file input functions.

  void SetMeshTitle(const char* title) { SetTitle(title); }
  void SetMeshDescription(const char* desc) { SetDescription(desc); }

  virtual ~GeneralMesh3f(void) {}

  INT4m GetSize() const { return mesh.GetSize();  }
  INT4m GetRange(BoundingBox &myrange) const {
    myrange=range; return 0;
  }
  INT4m GetBoundaryList(List<Vec3f> &boundary_list) const {
    boundary_list=boundary;
    return 0;
  }
  REAL4m GetApproximateCellWidth(void) const;
  Vec3f  GetApproximateCellDimensions(void) const;
  REAL4m GetPreferredStepSize(REAL4m step) const {
    REAL4m cellsize=GetApproximateCellWidth();
    return cellsize*(rint(step/cellsize));
  }
  REAL4m GetSubsampleGrit(void) const { return 0.1; }
  /// This is suppose to be a sensible lower bound on the sampling
  /// period relative to "GetApproximateCellWidth()".  This should
  /// probably be set to the smallest neighbor distance in the
  /// mesh, but for now we'll just use 0.1.
  INT4m GetDisplayList(REAL4m &xstep_request,
		       REAL4m & ystep_request,
		       REAL4m & zstep_request,
		       const BoundingBox &range_request,
		       List<LocatedVector> &display_list);

  ////////////// GeneralMesh3f specific functions ///////////////////
  void ResetMesh(); // Resets data w/o changing filename, title or desc.
  void ResetMesh(const char* newFilename); // Resets data & filename
  void ResetMesh(const char* newFilename,const char* newTitle,
		 const char* newDescription);
  /// Resets data, filename, title and description.

  void AddPoint(const LocatedVector& lv);
  void SortPoints();   // Post point import processing, for efficient
  // accessing.  Will be called automatically by GetDisplayList() if needed,
  // but client access provided because this may take more than a few CPU
  // cycles.
  void SetBoundaryList(const List<Vec3f> &boundary_list);
  void SetApproximateStepHints();
  void SetStepHints(const Vec3f &_step) { step_hints=_step; }
  Vec3f GetStepHints() const { return step_hints; }
};

#endif  // MESH_H
