/* FILE: err.cc                 -*-Mode: c++-*-
 *
 * Error handling
 *
 */

#undef CODE_CHECKS

#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "misc.h"

#if defined __alpha__ && defined __GNUC__
// On Linux/AXP, the 'swab' declaration is buried under
// a bunch of #ifdef's in <unistd.h>.  So this is just
// a stupid kludge to get it defined.
extern "C" {
extern void swab(const void *from, void *to, size_t n);
}
#endif

#if defined OLD_CODE && defined __GNUC__
extern "C" {
extern void swab(const void *from, void *to, size_t n);
char *strcat(char *dest, const char *src);
}
#endif


const char* StandardInput     = "Read input from stdin";
const char* StandardOutput    = "Write output to stdout";

const char* ErrNoMem          = "Insufficent working memory";
const char* ErrArrOutOfBounds = "Array index out of bounds";
const char* ErrBadParam       = "Bad parameter";
const char* ErrBadFile        = "File error";
const char* ErrBadFileOpen    = "Unable to open file";
const char* ErrBadFileOpenS   = "Unable to open file %s";
const char* ErrBadFileHeader  = "Bad file header";
const char* ErrMisc           = "Miscellaneous error";
const char* ErrXInit          = "Unable to initialize X display";
const char* ErrProgramming    = "Programming error";

#define BELL 0x07

// Greatest common divisor via Euclid's algorithm
int gcd(int m,int n)
{
  int temp;
  while((temp=(m%n))>0) { m=n; n=temp; }
  return n;
}

//  Fatal error handling routine
void fatalerr(const char *function_name,const char *message_fmt,...)
{
  va_list arg_ptr;
  fflush(stdout);

  if(function_name==(char *)NULL) function_name="<NULL>";
  if(message_fmt==(char *)NULL)   message_fmt="<NULL>";

  /* Print prefix */
  if(function_name[strlen(function_name)-1]!=')')
    fprintf(stderr,"%c\nFATAL ERROR in %s(): ",BELL,function_name);
  else
    fprintf(stderr,"%c\nFATAL ERROR in %s: ",BELL,function_name);
  
  /* Print message */
  va_start(arg_ptr,message_fmt);
  /* arg_ptr now points to first argument after message format string */
  vfprintf(stderr,message_fmt,arg_ptr);
  va_end(arg_ptr);
  
  /* Exit */
  fprintf(stderr,"\n%c\n",BELL);
  fflush(stderr);
  exit(1);
}

//  Non-fatal error handling routine
void nonfatalerr(const char *function_name,const char *message_fmt,...)
{
  va_list arg_ptr;
  fflush(stdout);
  static int call_count;
  
  if(function_name==(char *)NULL) function_name="<NULL>";
  if(message_fmt==(char *)NULL)   message_fmt="<NULL>";

  /* Print prefix */
  if(call_count<5) fprintf(stderr,"%c",BELL);
  if(function_name[strlen(function_name)-1]!=')')
    fprintf(stderr,"\nWarning from %s(): ",function_name);
  else
    fprintf(stderr,"\nWarning from %s: ",function_name);
  
  /* Print message */
  va_start(arg_ptr,message_fmt);
  /* arg_ptr now points to first argument after message format string */
  vfprintf(stderr,message_fmt,arg_ptr);
  va_end(arg_ptr);
  
  /* Return */
  if(call_count<5)  fprintf(stderr,"%c\n",BELL);
  else              fprintf(stderr,"\n");
  fflush(stderr);
  call_count++;
}

// Signal logging (Used for debugging)
#if defined(sun) && !defined(__svr4__) && !defined(__SVR4)
void LogSignal(int sig,...)
#else
void LogSignal(int sig)
#endif
{ // Signal logger
  fprintf(stderr,"Caught signal #%d\n",sig);
  signal(sig,LogSignal);
}
void SetupSignalLogger(void)
{
  for(int i=0;i<32;i++) signal(i,LogSignal);
}

// Terminal I/O
static struct termios tio_orig,tio_current;

#if defined(sun) && !defined(__svr4__) && !defined(__SVR4)
void CtrlZTerm(int,...)
#else
void CtrlZTerm(int)
#endif
{ // Ctrl-Z handler
  signal(SIGTSTP,CtrlZTerm);
  ResetTerm(0);
  kill(getpid(),SIGSTOP);  // Is this right???
}

#if defined(sun) && !defined(__svr4__) && !defined(__SVR4)
void ContTerm(int,...)
#else
void ContTerm(int)
#endif
{ // SIGCONT (continue after SIGSTOP) handler
  signal(SIGCONT,ContTerm);
  tcsetattr(fileno(stdin),TCSANOW,&tio_current);
}

void SetupTerm(void)
{
  // Save original tty state and set up signal handlers
  static int first_call=1;
  if(first_call==1) {
    // Save tty on first call
    tcgetattr(fileno(stdin),&tio_orig);
    first_call=0;
  }
  signal(SIGINT,ResetTerm);  signal(SIGQUIT,ResetTerm);
  signal(SIGTSTP,CtrlZTerm); signal(SIGCONT,ContTerm);
//  atexit(ResetTerm);  // This causes problems from forked children
}

void RawTerm(void)
{ // Set one character at a time input mode
  tio_current=tio_orig;
  tio_current.c_cc[VMIN] = 1; // Return when buffer has 1 or more characters
  tio_current.c_cc[VTIME] = 0; // No timeout (wait forever)
  tio_current.c_lflag &= ~ICANON;  // Turn off line input mode
  tcsetattr(fileno(stdin),TCSANOW,&tio_current);
}

void CookedTerm(void)
{ // Set up line-by-line input mode
  tio_current.c_lflag |= ICANON;
  tcsetattr(fileno(stdin),TCSANOW,&tio_current);
}

void RawWaitTermInput(void)
{ // This routine sleeps until input is available on stdin
  struct timeval timeout;
  fd_set rfds;
  while(1) {
    timeout.tv_sec=5; /* Seconds */
    timeout.tv_usec=0; /* Microseconds */
    /* Set up fd_set's for select() */
    FD_ZERO(&rfds); // Initialize rfds
    FD_SET(fileno(stdin),&rfds);  // Input ready?
    if(select(1,&rfds,
	      NULL,    // No writes
	      NULL,    // No exceptions
	      &timeout) == -1) break; // ERROR!
    if(FD_ISSET(fileno(stdin),&rfds)) break; // Character ready
  }
}

int TermInputReady(void)
{ // Returns 1 if at least one character is ready to be read; 0 otherwise
  struct timeval timeout;
  timeout.tv_sec=0; /* Seconds */
  timeout.tv_usec=0; /* Microseconds */

  /* Set up fd_set's for select() */
  fd_set rfds;
  FD_ZERO(&rfds); // Initialize rfds
  FD_SET(fileno(stdin),&rfds); // Input ready?

  if(select(1,&rfds,
	    NULL,    // No writes
	    NULL,    // No exceptions
	    &timeout) == -1) return 0; // ERROR!

  if(FD_ISSET(fileno(stdin),&rfds)) {
    return 1; // Character ready
  }
  return 0;
}

#if defined(sun) && !defined(__svr4__) && !defined(__SVR4)
void ResetTerm(int,...)
#else
void ResetTerm(int)
#endif
{ // Return tty to original state
  tcsetattr(fileno(stdin),TCSANOW,&tio_orig);
}

// Byte reversal routines;  Converts between big-endian
// and little-endian representations.  NOTE: These routines
// assume a word is 32 bits long, and a short int is 16 bits.
// ALSO, _ASSUMES_ a double is 64 bits long, but I don't know
// any good way to check this at compile time. :-(
#if !(defined USHRT_MAX && defined UINT_MAX)
#include <limits.h>
#endif
#if ( USHRT_MAX != 65535U ) // Should be #define'd in <limits.h>
#error "Byte/word reversal code assumes short int's are 16 bits long."
#endif
#if ( UINT_MAX != 4294967295U ) // Should be #define'd in <limits.h>
#error "Byte/word reversal code assumes words (int's) are 32 bits long."
#endif

unsigned short flip(unsigned short input)
{
  unsigned char byte1,byte2;
  byte1=input/0x100;   byte2=input&0xFF;
  return ((unsigned short)byte2)*0x100+byte1;
}

short flip(short input)
{
  return (short)flip((unsigned short)input);
}

unsigned int flip(unsigned int input)
{
  unsigned short word1,word2;
  word1=flip((unsigned short)(input/0x10000));
  word2=flip((unsigned short)(input&0xFFFF));
  return ((unsigned int)word2)*0x10000+word1;
}

int flip(int input)
{
  return (int)flip((unsigned int)input);
}

double flip(double input)
{
  unsigned int *uiptr,temp;
  uiptr=(unsigned int *)(&input);
  temp = flip(*uiptr);
  *uiptr = flip(*(uiptr+1));
  *(uiptr+1) = temp;
  return input;
}

void swap_bytes(void *buf,int buflen)
{
  swab((const char *)buf,(char *)buf,buflen);
}

void swap_words(void *buf,int count)
{
  unsigned int *bufptr;
  bufptr=(unsigned int *)buf;
  for(;count>0;bufptr++,count--) (*bufptr)=flip(*bufptr);  
}

void swap_doublewords(void *buf,int count)
{
  unsigned int *uiptr,temp;
  swap_words(buf,2*count);
  for(uiptr=(unsigned int *)buf;count>0;uiptr+=2,count--) {
    temp= *uiptr; *uiptr= *(uiptr+1); *(uiptr+1)=temp;
  }
}


// Random number generator
double UnifRand(unsigned int seed)
{ // Returns random value in [0,1] with unif. distrib.
  // seed>0 reinitializes the generator with "seed"

  static int first=1;

  if(seed>0 || first) {
    int exercise=100;
    if(seed==0) { // Use random seed
      struct timeval tv;
      gettimeofday(&tv,NULL);
      seed=tv.tv_sec+tv.tv_usec;
      exercise+=tv.tv_usec%97;
    }
    srandom(seed);
    for(first=0;first<exercise;first++) random();  // Exercise generator
    first=0;
  }
  return ((double)random())/((double)RANDOM_MAX);
}

// Timer routines
void TrackTime::Reset(void)
{ // Initializes TrackTime private variables (start times)
  // NOTE: We should replace 'getrusage' calls with calls to
  //       'times' (2), since the latter is POSIX conforming.
  //       (getrusage is a BSD-ism.)
  getrusage(RUSAGE_SELF,&startusage);
  gettimeofday(&starttime,NULL);
}

TrackTime::TrackTime(void)
{
  Reset();
}

char* TrackTime::FormatDiff(char* str,
			   const struct timeval &t1,const struct timeval &t2)
{ // Formats |t2-t1| into str in format HH:MM:SS.SS, so strlen(str)>=12
  // if HH<=99, and >12 if HH>99.
  long hour,min,sec,hunsec,work;

  work=t2.tv_sec-t1.tv_sec;
  hunsec=t2.tv_usec-t1.tv_usec;

  if(work<0 || (work==0 && hunsec<0)) { // t1 is later than t2; swap
    work*=-1; hunsec*=-1;
  }

  if(hunsec<0) { // Need carry from tv_sec
    hunsec+=1000000;
    work--;
  }
  hunsec=(hunsec+5000)/10000; // Round to nearest hundredth of a second
  while(hunsec>=100) { work++; hunsec-=100; } // In case we rounded up

  hour=work/3600;  work-=hour*3600;
  min=work/60;     work-=min*60;
  sec=work;

#ifdef CODE_CHECKS
  if(hour<0 || hour>99 || min<0 || min>59 || sec<0 || sec>59) {
    fprintf(stderr,"Hmm, there seems to be a timer problem\n");
    fprintf(stderr,"t1.sec=%d, t1.usec=%d\n",t1.tv_sec,t1.tv_usec);
    fprintf(stderr,"t2.sec=%d, t2.usec=%d\n",t2.tv_sec,t2.tv_usec);
  }
#endif // CODE_CHECKS

  sprintf(str,"%02ld:%02ld:%02ld.%02ld",hour,min,sec,hunsec);
  return str;
}


#ifdef __svr4__
char* TrackTime::FormatDiff(char* str,const struct timestruc_t &t1,
			    const struct timestruc_t &t2)
{
  struct timeval tt1,tt2;
  tt1.tv_sec=t1.tv_sec;
  tt1.tv_usec=(t1.tv_nsec+500)/1000; // Convert from ns to us
  tt2.tv_sec=t2.tv_sec;
  tt2.tv_usec=(t2.tv_nsec+500)/1000; // Convert from ns to us
  return FormatDiff(str,tt1,tt2);
}
#endif // __svr4__

void TrackTime::Print(FILE *fptr)
{ // Prints process run time information to *fptr
  struct timeval curtime;
  struct rusage usage;
  char ct[32];
  getrusage(RUSAGE_SELF,&usage);
  gettimeofday(&curtime,NULL);
  fprintf(fptr,"##############################\n");
  fprintf(fptr,"# User    time = %s #\n",
	  FormatDiff(ct,startusage.ru_utime,usage.ru_utime));
  fprintf(fptr,"# System  time = %s #\n",
	  FormatDiff(ct,startusage.ru_stime,usage.ru_stime));
  fprintf(fptr,"# Elapsed time = %s #\n",
	  FormatDiff(ct,starttime,curtime));
  fprintf(fptr,"##############################\n");
}

void TrackTime::PrintUser(FILE *fptr)
{ // Prints process user run time information to *fptr
  struct rusage usage;
  char ct[32];
  getrusage(RUSAGE_SELF,&usage);
  fprintf(fptr,"User time: %s\n",
	  FormatDiff(ct,startusage.ru_utime,usage.ru_utime));
}

StopWatch::StopWatch(void)
{
  useduser.tv_sec=usedsystem.tv_sec=usedelapsed.tv_sec=0;
  useduser.tv_usec=usedsystem.tv_usec=usedelapsed.tv_usec=0;
  runstatus=0;
}

timeval StopWatch::Diff(timeval t2,timeval t1)
{ // Performs t2-t1
  timeval t3;
  t3.tv_sec=t2.tv_sec-t1.tv_sec;
  t3.tv_usec=t2.tv_usec-t1.tv_usec;

  if(t3.tv_usec<0) { // Need carry from tv_sec
    t3.tv_usec+=1000000;
    t3.tv_sec--;
  }
  return t3;
}

void StopWatch::Update(timeval &t1,timeval t2)
{ // Performs t1+=t2
  t1.tv_sec+=t2.tv_sec;
  if((t1.tv_usec+=t2.tv_usec)>1000000) {
    t1.tv_usec-=1000000;
    t1.tv_sec++;
  }
}

void StopWatch::Start(void)
{
  if(runstatus!=0) return;  // Clock already running
  getrusage(RUSAGE_SELF,&startusage);
  gettimeofday(&starttime,NULL);
  runstatus=1;
}

void StopWatch::Stop(void)
{
  if(runstatus==0) return;  // Clock already stopped
  timeval stoptime;  rusage  stopusage;
  getrusage(RUSAGE_SELF,&stopusage);  gettimeofday(&stoptime,NULL);
  Update(useduser,Diff(stopusage.ru_utime,startusage.ru_utime));
  Update(usedsystem,Diff(stopusage.ru_stime,startusage.ru_stime));
  Update(usedelapsed,Diff(stoptime,starttime));
  runstatus=0;
}

char* StopWatch::FormatTime(char* str,struct timeval t)
{
  long hour,min,sec,hunsec;

  hunsec=(t.tv_usec+5000)/10000;  // Round to nearest hundredth of a second
  sec=t.tv_sec;
  while(hunsec>=100) { sec++; hunsec-=100; } // In case we rounded up & over
  hour=sec/3600;  sec-=hour*3600;
  min=sec/60;     sec-=min*60;

  sprintf(str,"%02ld:%02ld:%02ld.%02ld",hour,min,sec,hunsec);
  return str;
}

void StopWatch::Print(FILE *fptr)
{
  char timestr[32];

  if(runstatus!=0) { // Clock running;
    Stop(); // Quickly stop & start to update used times
    Start();
  }

  fprintf(fptr,"######### StopWatch ##########\n");
  fprintf(fptr,"# User    time = %s #\n",
	  FormatTime(timestr,useduser));
  fprintf(fptr,"# System  time = %s #\n",
	  FormatTime(timestr,usedsystem));
  fprintf(fptr,"# Elapsed time = %s #\n",
	  FormatTime(timestr,usedelapsed));
  fprintf(fptr,"##############################\n");
}

// Wall clock time
void WallClock(FILE *fptr)
{
  time_t secs;
  struct timeval curtime;
  gettimeofday(&curtime,NULL);
  secs=(time_t)curtime.tv_sec;
  tm *expanded_time=localtime(&secs);
  fprintf(fptr,"%02d:%02d:%02d.%03d",
	  expanded_time->tm_hour,expanded_time->tm_min,
	  expanded_time->tm_sec,curtime.tv_usec/1000);
}

// Memory tracking routines
void TrackMem::Print(void)
{ // Prints current memory usage information to stdout;
  // Note 1: Explore "popen(3)" to redirect output.
  // Note 2: getrusage() doesn't seem to set memory fields in Linux 1.3.21
  // Note 3: The environment variable $PPID is only set under bash
  // Note 4: SGI's have a funny ps
  char cmd[128];
#ifdef sgi
  sprintf(cmd,"ps -l -p %d",getpid());
#else // Not SGI
  sprintf(cmd,"ps -m %d",getpid());
#endif
  system(cmd);
}


// Silly little routines to allocate/deallocate a doubly indexed array of
// doubles.
double** AllocateDoubles(int dim1,int dim2)
{
  double **arr;
  if(!(arr=new double*[dim1]))
    fatalerr("AllocateDoubles","Insufficient memory to allocate %d double*'s",
	     dim1);
  for(int i=0;i<dim1;i++) {
    if (!(arr[i]=new double[dim2])) fatalerr("AllocateDoubles",
	     "Insufficient memory to allocate %dx%d double's",dim1,dim2);
  }
  return arr;
}

void DeallocateDoubles(double** arr,int dim1)
{
  while(dim1>0) delete arr[--dim1];
  delete[] arr;
}

// Function to collapse runs of whitespace into single blanks.
// Also removes leading and trailing whitespace.
void CollapseStr(char* str)
{
  int i1,i2,lastblank;  char ch;
  for(i1=i2=0,lastblank=1;str[i2]!='\0';i2++) {
    if(isspace(ch=str[i2])) {
      if(!lastblank) { str[i1++]=' '; lastblank=1; }
    }
    else { str[i1++]=ch; lastblank=0; }
  }
  if(i1>0 && str[i1-1]==' ') str[i1-1]='\0';  // Remove trailing whitespace
  else                       str[i1]='\0';
}

// Function to transform an entire string into lowercase
void LowerStr(char* str)
{
  while(*str != '\0') { *str=tolower(*str); str++; }
}

// String class with automatic allocation and deallocation
void AutoString::CleanUp(void)
{
  if(size>0 && str!=(char *)NULL) delete str;
  size=0; str=(char *)NULL;
}

void AutoString::Init(int newsize)
{
  if(newsize!=size) {
    CleanUp();
    if(!(str=new char[newsize])) fatalerr("AutoString::Init",ErrNoMem);
    size=newsize;
  }
  str[0]=str[size-1]='\0'; // Safety
}

int AutoString::Dup(const char* newstr)
{
  if(newstr==str) return size;
  Init(strlen(newstr)+1);
  if(size>0) {
    int i=0; while(i<size-1) str[i++]= *(newstr++);
    str[size-1]='\0'; // Safety
  }
  return size;
}

int AutoString::Cat(int count,const char** strarr)
{
  int newsize,i;
  for(i=0,newsize=1;i<count;i++) {
    if(strarr[i]==str)
      fatalerr("AutoString::Cat(int,const char**)",
	       "strarr includes pointer to this->str");
    newsize+=strlen(strarr[i]);
  }
  Init(newsize);
  for(i=0,newsize=1;i<count;i++) strcat(str,strarr[i]);
  str[size-1]='\0'; // Safety
  return size;
}

int AutoString::Cat(int count,const char** strarr,const char* pad)
{
  int newsize,i;
  for(i=0,newsize=1;i<count;i++) {
    if(strarr[i]==str)
      fatalerr("AutoString::Cat(int,const char**,const char*)",
	       "strarr includes pointer to this->str");
    newsize+=strlen(strarr[i]);
  }
  newsize+=(count-1)*strlen(pad); // Leave room for pad string
  Init(newsize);
  strcpy(str,strarr[0]);
  for(i=1;i<count;i++) { strcat(str,pad); strcat(str,strarr[i]); }
  str[size-1]='\0'; // Safety
  return size;
}

int AutoString::Append(const char* addstr)
{
  // Get information and handle degenerate cases
#ifdef CODE_CHECKS
  if(addstr==str)
    nonfatalerr("AutoString::Append",
		"addstr==str;  This routine delete[]'s addstr, "
		"which might be something you don't want.");
#endif // CODE_CHECKS
  int addlen,newsize;
  addlen=strlen(addstr);
  if(addlen==0) return size; // addstr is empty
  if(size==0) return Dup(addstr); // Current str is empty
  newsize=size+addlen;  // Note: size includes terminating '\0'.

  // Allocate space for extended string
  char *newstr;
  if(!(newstr=new char[newsize])) fatalerr("AutoString::Append",ErrNoMem);

  // Copy old str plus addstr into new space
  int i; char *cptr=newstr;
  for(i=0;i<size-1;i++) *(cptr++)=str[i];
  for(i=0;i<addlen;i++) *(cptr++)=addstr[i];
  *cptr='\0';

  // Delete old str and make tempstr current
  CleanUp();
  str=newstr;
  size=newsize;
  str[size-1]='\0'; // Safety
  return size;
}

AutoString::AutoString(void)
{
  size=0; str=(char *)NULL;
}

AutoString::AutoString(const char* newstr)
{
  size=0; str=(char *)NULL;
  Dup(newstr);
}

AutoString::AutoString(const AutoString &a)
{
  size=0; str=(char *)NULL;
  Dup(a.str);
}

AutoString::~AutoString(void)
{
  CleanUp();
}

char AutoString::operator[](int n)
{
  if(n<0 || n>=size) fatalerr("AutoStr::operator[]",ErrArrOutOfBounds);
  return str[n];
}

