/* FILE: err.h                 -*-Mode: c++-*-
 *
 * Error handling
 *
 */

#ifndef ERR_H
#define ERR_H

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <unistd.h>

/* End includes */

#define GLOBAL_CODE_CHECK_LEVEL 0 
// Global control over file CODE_CHECKS macros:
// Values:  -1 => Disable all checks
//           0 => File discretionary checks
//         100 => Enable all checks
// NOTE: For this to work, all files must include this file (err.h)
//       before any other code.

// A few machine specific #define's
typedef int INT32;
typedef char BYTE;
typedef double FLOAT64;

typedef char* char_ptr;
typedef int*  int_ptr;
typedef double* double_ptr;

#undef OLDSPRINTF
#ifdef sgi
#include <bstring.h>
#define STUPIDDOLLARSIGNS
#else
// #if CC$gfloat
// #undef STUPIDDOLLARSIGNS
// #else
#define STUPIDDOLLARSIGNS
// #endif
#endif
#ifdef sun
#if defined(__svr4__) || defined(__SVR4)
//#include <sys/rusage.h>
#include <sys/resource.h>
// For _some_ reason, the following are missing from the Sun header files:
extern "C" {
int getrusage(int who, struct rusage *rusage); // Missing from sys/resource.h
int usleep(unsigned int useconds);  // Missing from unistd.h
long random();
int srandom( unsigned seed);
}
#else
#define XFilterEvent(xev,window) (0)
#define OLDSPRINTF
#endif
#endif

// Terminal I/O
#include <X11/keysym.h>
#ifndef XK_Page_Up             /* For Suns */
#define XK_Page_Up XK_Prior
#endif
#ifndef XK_Page_Down             /* For Suns */
#define XK_Page_Down XK_Next
#endif

#ifndef HAS_USLEEP
#define HAS_USLEEP 1   // #define this to 0 for systems without usleep()
#endif

#ifndef STUPIDDOLLARSIGNS 
#define HUGE_VAL 8.988465674311578540726371186585e+307
#else
#define EPSILON        2.22045e-16
#define SQRT_EPSILON   1.49012e-08
#ifndef PI
#define PI             3.14159265358979300
#endif
#endif

// Global physical constants
const double mu0=PI*4e-7;         // T.m/A

// Extended key codes
#define Alt_a  ((unsigned int)(0x80+'a'))
#define Alt_b  ((unsigned int)(0x80+'b'))
#define Alt_c  ((unsigned int)(0x80+'c'))
#define Alt_d  ((unsigned int)(0x80+'d'))
#define Alt_e  ((unsigned int)(0x80+'e'))
#define Alt_f  ((unsigned int)(0x80+'f'))
#define Alt_g  ((unsigned int)(0x80+'g'))
#define Alt_h  ((unsigned int)(0x80+'h'))
#define Alt_i  ((unsigned int)(0x80+'i'))
#define Alt_j  ((unsigned int)(0x80+'j'))
#define Alt_k  ((unsigned int)(0x80+'k'))
#define Alt_l  ((unsigned int)(0x80+'l'))
#define Alt_m  ((unsigned int)(0x80+'m'))
#define Alt_n  ((unsigned int)(0x80+'n'))
#define Alt_o  ((unsigned int)(0x80+'o'))
#define Alt_p  ((unsigned int)(0x80+'p'))
#define Alt_q  ((unsigned int)(0x80+'q'))
#define Alt_r  ((unsigned int)(0x80+'r'))
#define Alt_s  ((unsigned int)(0x80+'s'))
#define Alt_t  ((unsigned int)(0x80+'t'))
#define Alt_u  ((unsigned int)(0x80+'u'))
#define Alt_v  ((unsigned int)(0x80+'v'))
#define Alt_w  ((unsigned int)(0x80+'w'))
#define Alt_x  ((unsigned int)(0x80+'x'))
#define Alt_y  ((unsigned int)(0x80+'y'))
#define Alt_z  ((unsigned int)(0x80+'z'))

#define RANDOM_MAX (2*((1<<30)-1)+1)

#define SQ(x) ((x)*(x))
#define MIN(x,y) ((x)<=(y)?(x):(y))
#define MAX(x,y) ((x)>=(y)?(x):(y))
#define ROUND(x) ((int)rint(x))

int gcd(int m,int n);  /* Greatest Common Divisor */

extern const char* StandardInput;
extern const char* StandardOutput;

extern const char* ErrNoMem;
extern const char* ErrArrOutOfBounds;
extern const char* ErrBadParam;
extern const char* ErrBadFile;
extern const char* ErrBadFileOpen;
extern const char* ErrBadFileOpenS;
extern const char* ErrBadFileHeader;
extern const char* ErrMisc;
extern const char* ErrXInit;
extern const char* ErrProgramming;

void fatalerr(const char *function_name,const char *message_fmt,...);
void nonfatalerr(const char *function_name,const char *message_fmt,...);

#if defined(sun) && !defined(__svr4__) && !defined(__SVR4)
void LogSignal(int sig,...);
void ResetTerm(int,...);
#else
void LogSignal(int sig);
void ResetTerm(int);
#endif
void SetupSignalLogger(void);

void SetupTerm(void);
void RawTerm(void);
void CookedTerm(void);
void RawWaitTermInput(void);
int TermInputReady(void);

// Byte reversal routines, for conversion between big-endian
// and little-endian representations
unsigned short flip(unsigned short input);
unsigned int flip(unsigned int input);
int flip(int input);
double flip(double input);
void swap_bytes(void *buf,int buflen);
void swap_words(void *buf,int count);
void swap_doublewords(void *buf,int count);

// Random number generator
double UnifRand(unsigned int seed=0);
// Returns random value in [0,1] with unif. distrib.
// seed>0 => reinitialize generator with "seed"

// Timer routines
class TrackTime {
private:
  struct timeval starttime;
  struct rusage  startusage;
  char* FormatDiff(char* str,
		   const struct timeval &t1,const struct timeval &t2);
#ifdef __svr4__
char* TrackTime::FormatDiff(char* str,const struct timestruc_t &t1,
			    const struct timestruc_t &t2);
#endif // __svr4__
public:
  TrackTime(void);
  void Reset(void);
  void Print(FILE *fptr);   // Prints process run times to file *fptr
  void PrintUser(FILE *fptr); // Prints user run time only
};

class StopWatch {
private:
  int runstatus; // 0=> Timing paused, 1=> Time running
  timeval starttime;
  rusage  startusage;
  timeval useduser,usedsystem,usedelapsed;
  timeval Diff(timeval t2,timeval t1);  // Performs t2-t1
  void    Update(timeval &t1,timeval t2);  // Performs t1+=t2
  char* FormatTime(char* str,timeval t);
public:
  StopWatch(void);
  void Start(void);
  void Stop(void);
  void Print(FILE *fptr);
};

void WallClock(FILE *fptr);  // Prints wall clock time (w/o date)

// Memory tracking routines
class TrackMem {
public:
  void Print(void);   // Prints process memory usage to stdout
};

// Silly little routines to allocate/deallocate a doubly indexed array of
// doubles.
double** AllocateDoubles(int dim1,int dim2);
void DeallocateDoubles(double** arr,int dim1);


// Function to collapse runs of whitespace into single blanks.
// Also removes leading whitespace.
void CollapseStr(char* str);

// Function to transform an entire string into lowercase
void LowerStr(char* str);

// String class with automatic allocation and deallocation
class AutoString {
private:
  int size;
  char *str;
  void Init(int newlen);
  void CleanUp(void);
public:
  AutoString(void);
  AutoString(const char* newstr);
  AutoString(const AutoString &a);
  ~AutoString(void);
  int GetSize(void) { return size; }
  int Dup(const char* newstr);
  int Dup(const AutoString& a) { return Dup(a.str); }
  AutoString& operator=(const AutoString& a) { Dup(a); return *this; }
  int Cat(int count,const char** strarr);
  int Cat(int count,const char** strarr,const char* pad);
     // Concatenates all strings in strarr, replacing current string, if any.
  int Append(const char* addstr);
  char operator[](int n);
  const char* GetStr(void) { return (const char *)str; }
  void Free(void) { CleanUp(); }
};

#endif // .NOT. ERR_H
