# FILE: extension.tcl
#
# The Extension Manager loads new code into a Tcl interpreter in
# units called 'extensions'.  For each extension, there is a database
# record which holds all the data describing the extension which
# the Extension Manager needs to do its work.  Below, the class for
# those database records, 'em_extension' is defined.
#
# TODO:
#
# Add documentation fields

if {[string match {} [info procs em_extension]]} {
em_class em_extension {

    const common statusValues \
        {new registered unselected defined selected loaded}

    common dirtoextmap {[em_array new]}

    common pfxtoextmap {[em_array new]}

    common log {[
            global DEBUGLVL
            if {[info exists DEBUGLVL] && $DEBUGLVL} {
                em_log new -dateformat "" \
                        -title "[info nameofexecutable] application message log"
            } else {
                list em_nop
            }
            ]}

    common nullfile {[
            global tcl_platform
            if {![info exists tcl_platform(platform)]} {
                list /dev/null
            } else {
                switch $tcl_platform(platform) {
                    unix {
                        list /dev/null
                    }
                    windows {
                        list NUL:
                    }
                    default {
                        error "can't determine null file"
                    }
                }
            }
            ]}

    proc NullFile {} {
        return $nullfile
    }


    # This is private because we don't want the callers of
    # em_extension new to provide an incorrect value.
    const private variable directory 

    const public variable name {
        if {![regexp {^[a-z]+$} $name]} {
            error "Attempt to give the name '$name' to the extension 
	defined in $directory .

	Extension names must be all lowercase alphabetic characters."
        }
    }

    const public variable prefix {
        if {[catch {$pfxtoextmap lookup $prefix} others]} {
            # No extension using that prefix yet.  OK to use it.  
            # Register its use.
            $pfxtoextmap set [list $prefix [list $this]]
        } else {
            # $others is list of extensions already using this prefix.
            # All of them have the same name.  Check that I do too.
            set ext [lindex $others 0]
            if {[string match $name [$ext cget -name]]} {
                # My name is the same.  Join the list of extensions 
                # sharing this prefix, if not already on it.
                if {[lsearch -exact $others $this] < 0} {
                    lappend $others $this
                    $pfxtoextmap set [list $prefix $others]
                }
            } else {
                # My name is different --> Error!
                error "Extension '$name', defined in $directory, 
	wants to use '$prefix' as its prefix.

	Extension '[$ext cget -name]', defined in [$ext directory], 
        is already using the prefix '[$ext cget -prefix]'.

	Extension prefixes must be unique."
            }
        }
    }

    const public variable versionsFile = versions.tcl
    const public variable contentsFile = contents.tcl
    const public variable addendaFile = addenda.tcl
    const public variable develop = 0

    # Use a status variable instead of lots of booleans
    variable status = new
    variable contents
    variable versioninfo
    variable version
    variable cmdpkgs
    variable attemptingredefine = 0

  constructor {args} {
    # The directory associated with an extension is the one which contains
    # The script file (pkgIndex.tcl) which contains the call 
    # 'em_extension new'
    set directory [file join [pwd] [string trimright [file dirname \
        [info script]] ./]]
    # DOME: Make connections to GUI interface
    if {[catch {$dirtoextmap lookup $directory} msg]} {
      # Construct a default name for the extension based on the directory name
      regexp {^[a-z]+} [string tolower [file tail $directory]] name
      # No prefix is a good default for all extensions.
      # A default value is constructed from the extension name below.
      if {[info exists prefix]} {
        unset prefix
      }
      # Allow class user to set values of public variables (including constants)
      eval $this configure $args
      # If no extension prefix has been selected by the user, construct one
      # based on the extension name
      if {[info exists name] && ![info exists prefix]} {
        set pfx [string range $name 0 0]
        set rest [string range $name 1 end]
        regsub -all {[aeiou]} $rest {} rest
        set second [string range $rest 0 0]
        regsub -all $second $rest {} rest
        append pfx $second
        append pfx [string range $rest 0 0]
        # Use configure instead of direct set of value so that the 
        # value gets checked by the configuration script
        $this configure -prefix $pfx
      }
      $log Append "Constructing extension '$name' ($directory) ..."
      # No record for this directory is in the database yet.
      # Register this one.
      $dirtoextmap set [list $directory $this]
      set status registered

      #	* Be sure the files which define the extension's contents exist.
      #   If not and we're developing, create them as empty templates.
      #   	If unable (e.g. no write access) -> error
      #   If files don't exist and not developing -> installation error.
      #
      # contents file:
      set abscontentsFile [file join $directory $contentsFile]
      if {![file exists $abscontentsFile]} {
        if {$develop} {
          $class writeToFile $abscontentsFile \
"# Extension Manager contents file, version 0.0
# FILE: $contentsFile
#
# DO NOT EDIT THIS FILE!
#
# It is automatically created and modified by the extension manager
# package to maintain an official table of contents of this extension.
# 
# To add more commands to this extension as a local modification, edit
# the file ./$addendaFile .
#

\[em_extension lookup \[file join \[pwd\] \[string trimright \[file dirname \[info script\]\] ./\]\]\] contents {
}"
        } else {
          error "Missing contents file: $contentsFile
        in directory: $directory

	This file should have been included in the software distribution
	of the extension '$name'.  Check your installation."
        }
      }
      #
      # addenda file:
      set absaddendaFile [file join $directory $addendaFile]
      if {![file exists $absaddendaFile] && $develop} {
        $class writeToFile $absaddendaFile \
"# Extension Manager addenda file, version 0.0
# FILE: $addendaFile
#
# The conventional method for adding locally developed Tcl commands to 
# this extension:
#
# 1. Edit the file ./pkgIndex.tcl .  Add the option '-develop 1'
#    to the call to 'em_extension new'.  This causes the extension
#    to be defined in 'development mode,' a mode which includes
#    commands listed in this addenda file as well as those listed in the
#    official contents file.
#
# 2. Create a file in this directory which defines the new command
#    when sourced.  By convention, the name of the new file should be 
#    the same as the base name of the public command it defines, with 
#    a '.tcl' file extension.
#
# 3. Following the header of this file is a single Tcl command,
#    '\[em_extension lookup <...>\] contents <list>' where <list> is a 
#    list of files.  Add the name of the file created in step 2 to this 
#    list.
#
# 4. Repeat steps 2 and 3 for each new command to add to this extension.
#

\[em_extension lookup \[file join \[pwd\] \[string trimright \[file dirname \[info script\]\] ./\]\]\] contents {
}"
      }
      #
      # versions file:
      set absversionsFile [file join $directory $versionsFile]
      if {![file exists $absversionsFile]} {
        if {$develop} {
          $class writeToFile $absversionsFile \
"# Extension Manager versions file, version 0.0
# FILE: $versionsFile
#
# DO NOT EDIT THIS FILE!
#
# It is automatically created and modified by the extension manager
# package to maintain an official table of the current version of
# this extension and its command packages.
#

\[em_extension lookup \[file join \[pwd\] \[string trimright \[file dirname \[info script\]\] ./\]\]\] versioninfo {
.       0.0.0.0.0.0
}"
        } else {
          error "Missing versions file: $versionsFile
        in directory: $directory

	This file should have been included in the software distribution
	of the extension '$name'.  Check your installation."
        }
      }
      # Existence of definition files is assured.  Define this 
      # extension based on their contents:
      $this define
    } else {
      # This directory already has a record in the extension database.
      # Destroy this record.
      $this delete
      # Define the existing one, if necessary, and return it.
      if {[lsearch -exact $statusValues [$msg status]] < \
          [lsearch -exact $statusValues defined]} {
        $msg define
      }
      return $msg
    }
  }

  destructor {
    # Protection against failed construction
    if {[info exists status]} {
      # Undefine the extension, if necessary
      if {[lsearch -exact $statusValues $status] > \
          [lsearch -exact $statusValues registered]} {
        $this undefine
      }
      # Remove record from databases, if necessary.
      if {[lsearch -exact $statusValues $status] > \
          [lsearch -exact $statusValues new]} {
        $dirtoextmap unset $directory
        set l [$pfxtoextmap lookup $prefix]
        if {[llength $l] == 1} {
          $pfxtoextmap unset $prefix
        } else {
          set i [lsearch -exact $l $this]
          lreplace $l $i $i
          $pfxtoextmap set [list $prefix $l]
        }
      }
    }
  }

  method define {} {
    # Assure that extension is only defined if it doesn't already
    # have a definition.
    if {[lsearch -exact $statusValues $status] >= \
        [lsearch -exact $statusValues defined]} {
      error "'\$instance define' called on an em_extension instance
	which is already defined.  You must call '\$instance undefine'
	to clear the existing definition first.

        Leave it to Tcl's package mechanism to call '\$instance define' 
	at the right time."
    }
    $log Append "Defining extension '$name' ($directory) ..."
    # Set status to 'defined'
    set status defined
    # Form the absolute file names of the definition files
    set abscontentsFile [file join $directory $contentsFile]
    set absaddendaFile [file join $directory $addendaFile]
    set absversionsFile [file join $directory $versionsFile]
    # Source the contents and addenda files to get list of 
    # command packages to add
    if [catch {source $abscontentsFile} msg] {
      $this undefine
      error "Error in Extension Manager contents file 
	$abscontentsFile :
	$msg"
    }
    if {$develop} {
      if [catch {source $absaddendaFile} msg] {
        $this undefine
        error "Error in Extension Manager addenda file 
	$absaddendaFile :
	$msg"
      }
    }
    # Source versions file to get version information
    if [catch {source $absversionsFile} msg] {
      $this undefine
      error "Error in Extension Manager versions file 
	$absversionsFile :
	$msg"
    }
    # Check whether the versions file is out of date compared 
    # to any of the source files.
    set vtime [file mtime $absversionsFile]
    set outofdate {}
    foreach cmdpkg $contents {
      set src [lindex $cmdpkg 0]
      set srcFile [file join $directory $src]
      if {[file exists $srcFile]} {
        set ftime [file mtime $srcFile]
        if {($vtime < $ftime) || [catch {$versioninfo lookup $src}]} {
          lappend outofdate $src
        }
      } else {
        if {$develop} {
          $this undefine
          error "File $srcFile is listed in contents file 
	$abscontentsFile, or addenda file 
	$absaddendaFile, but does not exist."
        } else {
          $this undefine
          error "File $srcFile is listed in contents file 
	$abscontentsFile, but does not exist."
        }
      }
    }
    set ctime [file mtime $abscontentsFile]
    if {$develop} {
      set atime [file mtime $absaddendaFile]
    }
    if {[llength $outofdate] || ($vtime < $ctime) \
        || ($develop && ($vtime < $atime))} {
      lappend outofdate .
    }
    if {[llength $outofdate]} {
      # The versions file is out of date with respect to the files
      # listed in $outofdate.
      if {$develop} {
        $log Append "Updating versions file $absversionsFile ..."
        foreach src $outofdate {
          if {![catch {$versioninfo lookup $src} oldvnumber]} {
            set parts [split $oldvnumber .]
            for {set i [llength $parts]} {$i < 6} {incr i} {
              lappend parts 0
            }
            set revision [lindex $parts 5]
            set parts [lreplace $parts 5 end [incr revision]]
            $versioninfo set "$src [join $parts .]"
          } else {
            $versioninfo set "$src 0.0.0.0.0.0"
          }
          $log Append "$src is now version [$versioninfo lookup $src]"
        }
        $log Append "Writing new versions file $absversionsFile..."
        set vf "# Extension Manager versions file, version 0.0
# FILE: $versionsFile
#
# DO NOT EDIT THIS FILE!
#
# It is automatically created and modified by the extension manager
# package to maintain an official table of the current version of
# this extension and its command packages.
# 

\[em_extension lookup \[file join \[pwd\] \[string trimright \[file dirname \[info script\]\] ./\]\]\] versioninfo {
"
        foreach src [$versioninfo names] {
          append vf "$src	[$versioninfo lookup $src]\n"
        }
        append vf "}"
        if {[catch {$class writeToFile $absversionsFile $vf} msg]} {
          $this undefine
          error $msg
        }
      } else {
        $this undefine
        error "The Extension Manager versions file 
	$absversionsFile
	is out of date.  Check your installation."
      }
    }
    # Versions file is up to date (for now)
    # Now we know the version of this extension:
    set version [$versioninfo lookup .]
    # Use it to make a final sanity check
    # Each (name, version) should refer to only one extension so
    # we can support extension loading by (name, version) via Tcl's
    # pacakge command.
    set pins [package ifneeded $name $version]
    if {[llength $pins]} {
      $this undefine
      set ext [lindex $pins 0]
      if {[catch {$ext class} otherclass] \
          || ![string match em_extension $otherclass]} {
        error "The directory $directory defines
	extension '$name', version $version.

	An extension known by that name and version has already been
	defined in this interpreter via a non-Extension Manager mechanism.

	Extension (name, version) pairs must uniquely identify an
	extension."
      } else {
        error "The directory $directory defines
	extension '$name', version $version.

	Extension '[$ext cget -name]', version [$ext cget -version], has already
	been defined by the directory [$ext directory].

	Extension (name, version) pairs must uniquely identify an
	extension."
      }
    }
    # OK, everything checks out.  
    # Fill up the list of cmdpkgs for this extension
    set cmdpkgs {}
    foreach cmdpkg $contents {
      set src [lindex $cmdpkg 0]
      if {[catch {
        set newcp [em_cmdpkg @extension $this @file $src]
        switch -- [llength $cmdpkg] {
          1 {
            $newcp define #version [$versioninfo lookup $src]
          }
          2 {
            set pkgbase [lindex $cmdpkg 1]
            $newcp define #version [$versioninfo lookup $src] #pkgbase $pkgbase
          }
          3 {
            set pkgbase [lindex $cmdpkg 1]
            set cmdbase [lindex $cmdpkg 2]
            $newcp define #version [$versioninfo lookup $src] \
                #pkgbase $pkgbase #cmdbase $cmdbase
          }
        }
      } msg]} {
        $log Append "New cmdpkg $src failure: $msg"
        $this undefine
        error $msg
      }
      if {[lsearch $cmdpkgs $newcp] < 0} {
        lappend cmdpkgs $newcp
      } else {
        if {$develop} {
          $this undefine
          error "The file '[file join $directory [$newcp file]]'
	is listed as the source for two command packages in the files
	'$abscontentsFile' and 
	'$absaddendaFile'.

	The mapping between command packages and files must be one to one."
        } else {
          $this undefine
          error "The file '[file join $directory [$newcp file]]'
	is listed twice in '$abscontentsFile'.  
	Check your installation."
        }
      }
    }
    # Record stored in database, make extension known to package command
    package ifneeded $name $version [list $this select]
  }

  method undefine {} {
    statusValues status name directory cmdpkgs contents versioninfo version log
  } {
    if {[lsearch -exact $statusValues $status] < \
        [lsearch -exact $statusValues defined]} {
      return
    }
    $this unselect
    $log Append "Undefining extension '$name' ($directory) ..."
    if {[info exists cmdpkgs]} {
      foreach cmdpkg $cmdpkgs {
        $cmdpkg undefine
      }
      unset cmdpkgs
    }
    if {[info exists contents]} {
      unset contents
    }
    # Destroy version information array
    if {[info exists versioninfo]} {
      $versioninfo delete
      unset versioninfo
    }
    if {[info exists version]} {
      unset version
    }
    set status registered
  }

  method select {} {statusValues status name version cmdpkgs log} {
    # The package require loads the auto_index to load a certain extension
    # based on name/version (by running appropriate ifneeded script).
    # Invocation of any command in the extension triggers load of whole
    # extension.  Select each cmdpkg too.
    #
    # Assure that extension is only selected if it hasn't already
    # been selected.
    if {[lsearch -exact $statusValues $status] >= \
        [lsearch -exact $statusValues selected]} {
      # Exception for the 'em' extension (possibly others?)
      if {[string match _em_extension0 $this]} {
        return $this
      }
      error "'\$instance select' called on em_extension instance $this ,
	which is already selected.  You must first call 
	'$this undefine' to clear the existing selection, and
        '$this define' to reset the database.

	Better still, call '$this undefine', then
	'package require $name $version' and leave it to Tcl's package 
	mechanism to call '\$instance select' at the right time."
    }
    if {[lsearch -exact $statusValues $status] < \
        [lsearch -exact $statusValues defined]} {
      error "'\$instance select' called on em_extension instance $this ,
	which has not been defined.  You must first call 
        '$this define' to setup the database.

	Better still, call 'package require $name $version' and leave 
	it to Tcl's package mechanism to call '\$instance select' at 
	the right time."
    }
    $log Append "Selecting extension '$name', version $version ..."
    set status selected
    global auto_index
    package provide $name $version
    # Fill in the auto_index array
    foreach cmdpkg $cmdpkgs {
      $cmdpkg select
      #upvar #0 $cmdpkg cv
      #set auto_index($cv(command)) [list $this load]
      set auto_index([$cmdpkg command]) [list $this load]
    }
    return $this
  }

  private method unselect {} {
    statusValues status name version cmdpkgs dirtoextmap log
  } {
    # It would be nice to make this a public method, but the Tcl
    # command 'package forget <packagename>' has two effects which 
    # cannot be separated: it forgets the selection of a version of
    # the named package made by a prior 'package provide' *and* it 
    # also forgets all package definition information associated with
    # that package name.  This undoes part of both our 'define' and
    # 'select' methods.  Since we need to call that routine here,
    # we need to be sure unselect and undefine are combined in one
    # operation from the user perspective.  That means independent access 
    # to this method is denied.  Users must call the 'undefine' method,
    # which calls this one if needed.
    #
    # A side effect of the implementation of 'package forget' is that
    # we have to undefine all other extensions which have the same name
    # also.
    if {[lsearch -exact $statusValues $status] <= \
        [lsearch -exact $statusValues defined]} {
      return
    }
    $this unload
    $log Append "Unselecting extension '$name', version $version ..."
    set status unselected
    global auto_index
    foreach cmdpkg $cmdpkgs {
      $cmdpkg unselect
      unset auto_index([$cmdpkg command])
    }
    foreach extdir [$dirtoextmap names] {
      set ext [$dirtoextmap lookup $extdir]
      if {[string match $name [$ext cget -name]]} {
        $ext undefine
      }
    }
    package forget $name
  }
 
  method load {} {
    statusValues status name version cmdpkgs develop log attemptingredefine
  } {
    # Load the cmdpkgs of this extension into the interpreter by 
    # invoking the load method of each one.
    #
    # Assure that extension is only loaded if it hasn't already
    # been loaded.
    if {[lsearch -exact $statusValues $status] < \
        [lsearch -exact $statusValues selected]} {
      # Multiple calls to load the em_extension
      if {[string match _em_extension0 $this]} {
        return $this
      }
      error "'\$instance load' called on em_extension instance $this ,
	which has not been selected.  You must first call 
        '$this select' to setup the auto-loader database.

	Better still, call 'package require $name $version' and leave 
	it to Tcl's package mechanism to call '\$instance select' at 
	the right time."
    }
    $log Append "Loading extension '$name', version $version ..."
    set status loaded
    #
    # loading an extension causes all its command packages to be loaded.
    #
    # Each 'package require' below causes the sourcing of the
    # corresponding 'package ifneeded' script (the load method
    # of that command package).
    foreach cmdpkg $cmdpkgs {
      upvar #0 $cmdpkg cv
      if {[catch [list package require $cv(name)] msg]} {
        # An error from 'package require' means that the package could
        # not be loaded, given its current definition.  
        # Sometimes redefinition of the extension is sufficient to recover
        # from the error (for example, when the error is that a file has been
        # modified after extension definition).  When the extension is in
        # development mode, and such circumstances exist, a second load
        # attempt should be made.
        if {$develop && [lsearch -exact $statusValues $status] >= \
            [lsearch -exact $statusValues defined] && !$attemptingredefine} {
        $log Append $msg
        $log Append "
	Trying redefinition and re-load..."
          $this undefine
          set attemptingredefine 1
          $this define
          $this select
          $this load
          set attemptingredefine 0
          break
        } else  {
          # The load error is not recoverable.
          # Undefine the extension and report the error.
          $this undefine
          error "Error loading command package:
	$msg"
        }
      }
    }
  }

  method unload {} {statusValues status name version cmdpkgs log} {
    if {[lsearch -exact $statusValues $status] <= \
        [lsearch -exact $statusValues selected]} {
      return
    }
    $log Append "Unloading extension '$name', version $version ..."
    set status selected
    foreach cmdpkg $cmdpkgs {
      $cmdpkg unload
    }
  }

  method directory {} {directory} {
    return $directory
  }

  method cmdpkgs {} {cmdpkgs} {
    return $cmdpkgs
  }

  method status {} {status} {
    return $status
  }

  method version {} {version} {
    return $version
  }

  method contents {{contentslist {}}} {contents} {
    # This method provides access to the private variable 'contents'.
    # It appends each element of its argument list to the private
    # variable 'contents'.
    if {![info exists contents]} {
      set contents {}
    }
    foreach tuple $contentslist {
      # DOME: Need to add syntax check here
      if {[llength $tuple] > 3} {
        error "Malformed entry in contents list: $tuple"
      }
      lappend contents $tuple
    }
  }

  method versioninfo {{keyvallist {}}} {versioninfo} {
    if {![info exists versioninfo]} {
      set versioninfo [em_array new]
    }
    $versioninfo set $keyvallist
  }

  proc writeToFile {filename contents} {} {
    global errorCode errorInfo
    set error [ catch {
      set f [open $filename w]
      puts $f $contents
      close $f
    } msg]
    if {$error} {
      set code $errorCode
      set info $errorInfo
      catch {close $f}
      error "Error writing file $filename: 
	$msg" $info $code
    }
  }

  proc lookup {keydir} {dirtoextmap} {
    if {[catch {$dirtoextmap lookup $keydir} ext]} {
      error "No extension is defined in the directory 
	$keydir

	Perhaps you forgot a call to 'package require'?
	Or perhaps a 'pkgIndex.tcl' file is missing?"
    }
    return $ext
  }

  proc ApplicationLog {} {log} {
    return $log
  }
}
}
