# FILE: mmGraph.tcl
#
# The network wrapping for an instance of the omf_graphwin class.
#

# Set up for auto-loading
set master(relhdir) [string trimright [file dirname [info script]] ./]
set master(abshdir) [file join [pwd] $master(relhdir)]
set master(pdir) [file dirname $master(abshdir)]
set master(libdir) [file dirname $master(pdir)]
if {[lsearch $auto_path $master(libdir)] < 0} {
    lappend auto_path $master(libdir)
}

# Support libraries
package require Tk
package require em
package require oommf
package require net

# Death
proc Die {} {
    global protocol
    if {[info exists protocol]} {
        $protocol delete
        after 2000 exit
    } else {
        exit
    }
}

# Menu bar
set menubar [frame .mb -borderwidth 2 -relief raised]

set file_select [menubutton $menubar.file -text "File" \
	-menu $menubar.file.menu -underline 0]
set filemenu [menu $file_select.menu -tearoff 0]
$filemenu add command -label "Exit" -command { Die } -underline 1
wm protocol . WM_DELETE_WINDOW { Die }

set xaxis_select [menubutton $menubar.xaxis -text "X-Axis" \
	-menu $menubar.xaxis.menu -underline 0]
set xmenu [menu $xaxis_select.menu -tearoff 1]

set yaxis_select [menubutton $menubar.yaxis -text "Y-Axis" \
	-menu $menubar.yaxis.menu -underline 0]
set ymenu [menu $yaxis_select.menu -tearoff 1]

set options_select [menubutton $menubar.options -text "Options" \
	-menu $menubar.options.menu -underline 0]
set optionsmenu [menu $options_select.menu -tearoff 0]
set showkey 1
$optionsmenu add checkbutton -label "Key" -underline 0 \
		-variable showkey -offvalue 0 -onvalue 1 \
		-command {$graph SetKeyState $showkey}
$optionsmenu add command -label "Reset" -command {$graph Reset} -underline 0
$optionsmenu add command -label "Rescale" -command {$graph Rescale} -underline 2

pack $file_select $xaxis_select $yaxis_select $options_select -side left
pack $menubar -side top -fill x

# Proc to add new labels to the x & y axis menus.  It is not currently
# possible to remove a label once added.  A global array menu_select
# stores the state of the axis menu select boxes:
#    menu_select(x)        = name of selected label, or {} if none
#    menu_select(y,$label) = 1 if box associated with label is selected,
#                            0 otherwise.
set menu_select(x) {}
proc AppendAxesSelect { new_labels } {
    global xmenu ymenu menu_select
    # Add new entries
    foreach label $new_labels {
	set menu_select(y,$label) 0
	regsub -all "\[ \r\t\n\]" $label { } compressed_label
	$xmenu add radiobutton -label $compressed_label \
		-variable menu_select(x) -value $label  \
		-command CreateGraph
	$ymenu add checkbutton -label $compressed_label \
		-variable menu_select(y,$label) \
		-offvalue 0 -onvalue 1 \
		-command "ChangeCurve \{$label\}"
    }
}

# Graph
set graph_frame [frame .gf -borderwidth 0]
set graph [omf_graphwin new $graph_frame]
$graph configure -autodraw 1 -showkey $showkey
pack [$graph cget -winpath] -side top -fill both -expand 1
pack $graph_frame -side top -fill both -expand 1
em_eventHandler new $graph delete { after 2000 exit }

# Proc to update graph with latest data input triples
proc UpdateGraph { index } {
    global menu_select data_value data_label graph
    if [string match {} $menu_select(x)] { return } ;# No abscissa selected
    set abscissa $menu_select(x)
    if {![info exists data_value($index,$abscissa)]} { return }
    ## No abscissa value at this index
    set xvalue $data_value($index,$abscissa)
    foreach ordinate $data_label(all) {
	if {![info exists menu_select(y,$ordinate)] || \
		!$menu_select(y,$ordinate) } { continue }  ;# Skip curve
	if {[info exists data_value($index,$ordinate)]} {
	    $graph AddPoints $ordinate $xvalue $data_value($index,$ordinate)
	}
    }
    return
}

# Proc to add or delete a curve from the graph, as determined by the
# value of menu_select(y,$ordinate) (1=>display, 0=>don't display).
proc ChangeCurve { ordinate } {
    global menu_select data_value data_label graph
    $graph DeleteCurve $ordinate ;# Remove any previously curve by this name

    if {![info exists menu_select(y,$ordinate)]} { return }  ;# Safety
    if {!$menu_select(y,$ordinate)} {
	# Curve turned off.  If this is last curve, delete y-label.
	set curve_count 0
	foreach curve $data_label(all) {
	    if {![info exists menu_select(y,$curve)] || \
		    !$menu_select(y,$curve) } { continue }  ;# Skip curve
	    incr curve_count
	}
	if {$curve_count==0} {
	    $graph SetYLabel {}
	    $graph DrawAxes
	}
	return
    }
    if [string match {} $menu_select(x)] { return 0 } ;# No abscissa selected
    set abscissa $menu_select(x)
    # If y-label is not yet set, build one from current curve.
    set ylabel [lindex [$graph GetLabels] 1]
    if {[string match {} $ylabel]} {
	if {[info exists data_label(unit,$ordinate)]} {
	    set yunit $data_label(unit,$ordinate)
	    if {[string match {} $yunit]} { set yunit "Non Dim" }
	    $graph SetYLabel "$yunit"
	    $graph DrawAxes
	}
    }
    # Write curve
    set data_list {}
    $graph NewCurve $ordinate
    for { set i 0 } { $i < $data_value(count) } { incr i } {
	if {[info exists data_value($i,$abscissa)] && \
		[info exists data_value($i,$ordinate)]} {
	    append data_list \
		    " $data_value($i,$abscissa) $data_value($i,$ordinate)"
	}
    }
    eval {$graph AddPoints $ordinate} $data_list
    return
}

# Proc to create a complete new graph, erasing any previously existing
# graph.  Return value is the number of curves in new graph.
proc CreateGraph {} {
    global menu_select data_value data_label graph
    $graph ResetGraph    ;# Remove any previously existing graph
    if [string match {} $menu_select(x)] { return 0 } ;# No abscissa selected
    set abscissa $menu_select(x)
    if {![info exists data_label(unit,$abscissa)]} { return 0 } ;# Safety
    set xunit " ($data_label(unit,$abscissa))"
    if {[string match " ()" $xunit]} { set xunit {} }
    $graph SetXLabel "$abscissa$xunit"
    set curve_count 0
    foreach ordinate $data_label(all) {
	if {![info exists menu_select(y,$ordinate)]} { continue } ;# Safety
	if {! $menu_select(y,$ordinate) } { continue }        ;# Skip curve
	ChangeCurve $ordinate
	incr curve_count
    }
    if { $curve_count<1 } { $graph DrawAxes }
    return $curve_count
}


# Data storage
#   Data is stored in two global arrays: data_value and data_label.
# The first is indexed by entry order and label, e.g.,
#                data_value(4,Bx)
# There is also an element data_value(count) which lists the number
# of entry indexes, i.e., the first number in the indexing pair.
#   The data_label array  associates a "units" value to each label.
# This is set only the first time a label is used.  Elements are
# notated as
#                 data_label(unit,$label)
#
# There is also a data_label(all) entry which is a list of all
# labels.  There is currently no way to remove a label once it
# has been added.
#   The return value is a list of all new labels (if any).
#
set data_value(count) 0
set data_label(all) {}
proc StoreDataTable { triples } {
    global data_value data_label
    set index $data_value(count)
    set new_labels {}
    foreach trip $triples {
	set label [lindex $trip 0]
	set unit  [lindex $trip 1]
	set value [lindex $trip 2]
	set data_value($index,$label) $value
	if {![info exists data_label(unit,$label)]} {
	    set data_label(unit,$label) $unit
	    lappend new_labels $label
	}
    }
    set data_label(all) [join [list $data_label(all) $new_labels]]
    incr data_value(count)
    UpdateGraph $index
    return $new_labels
}

# Protocol
set protocol [net_protocol new -name "OOMMF DataTable protocol 0.0"]
$protocol AddMessage start HasGui {} {
    return [list start [list 0 0]]
}
$protocol AddMessage start DataTable { triples } {
    set new_labels [StoreDataTable $triples]
    if {![string match {} $new_labels]} {
	# New labels added
	AppendAxesSelect $new_labels
    }
}

set server [net_server new -protocol $protocol -alias mmGraph -log em_nop]
$server Start 0
wm title . mmGraph<[pid]>

# Enter event loop
vwait omf_forever
